import json
from traceback import format_exc
from functools import partial
from flask import request, current_app, jsonify
from werkzeug.exceptions import default_exceptions
from devtools.web import merge_fields


def http_exception_handler(error, debug):
    code = getattr(error, 'code', 500)
    if str(code).startswith('5'):
        description = json.dumps(merge_fields(request), indent=4)
        current_app.logger.error(description, exc_info=True)
        if debug:
            print(description)
            print(format_exc())
    return jsonify(
        name=getattr(error, 'name', ''),
        description=getattr(error, 'description', ''),
        path=request.path), code


def common_exception_handler(error, debug):
    req = json.dumps(merge_fields(request), indent=4)
    current_app.logger.error(req, exc_info=True)
    if debug:
        print(req)
        print(format_exc())
    return jsonify(
        message='Something wrong...',
        error=str(error)), 500


class Errors:

    def __init__(self, app=None, debug=False):
        self.debug = debug
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        for code in default_exceptions:
            app.errorhandler(code)(partial(http_exception_handler, debug=self.debug))
        app.errorhandler(Exception)(partial(common_exception_handler, debug=self.debug))



from requests import Session


class JwtSession(Session):
    token = ''

    def __init__(self, server, username="lvda", password="secret",
                 auth_url='/auth'):
        self.server = server
        self.username = username
        self.password = password
        self.auth_url = server + auth_url
        super().__init__()

    def get_token(self):
        data = {
            "username": self.username,
            "password": self.password,
        }
        resp = super().request(
            'POST', self.auth_url, data=data)
        if resp.status_code == 200:
            self.token = resp.json()['access_token']
        else:
            raise ConnectionError

    def request(self, method, url, **kwargs):
        kwargs['headers'] = {"Authorization": f"JWT {self.token}"}
        url = self.server + url
        resp = super().request(method, url, **kwargs)
        if resp.status_code == 401:
            self.get_token()
            kwargs['headers']["Authorization"] = f"JWT {self.token}"
            resp = super().request(method, url, **kwargs)
        return resp




'''
    类似于 lru_cache, 但提供限时的缓存
'''
from datetime import datetime, timedelta
import functools


def timed_cache(maxsize=128, typed=False, **timedelta_kwargs):
    def _wrapper(f):
        update_delta = timedelta(**timedelta_kwargs)
        next_update = datetime.utcnow() - update_delta
        f = functools.lru_cache(maxsize, typed)(f)

        @functools.wraps(f)
        def _wrapped(*args, **kwargs):
            nonlocal next_update
            now = datetime.utcnow()
            if now >= next_update:
                f.cache_clear()
                next_update = now + update_delta
            return f(*args, **kwargs)
        return _wrapped
    return _wrapper


def counted_cache(count, maxsize=128, typed=False):
    '''
        When count is zero, cache was cleared each time function
        was called, so basically means no cache.
    '''
    def _wrapper(f):
        counter = 0
        f = functools.lru_cache(maxsize, typed)(f)

        @functools.wraps(f)
        def _wrapped(*args, **kwargs):
            nonlocal counter
            res = f(*args, **kwargs)
            if counter < count:
                counter += 1
            else:
                f.cache_clear()
                counter = 0
            return res
        return _wrapped
    return _wrapper


if __name__ == '__main__':
    import time

    @counted_cache(1)
    def test():
        return datetime.now()

    for i in range(10):
        time.sleep(0.1)
        print(i, test())



import re
from ipaddress import ip_address
from rest_framework.serializers import Field, ValidationError


class MegaByteField(Field):
    def to_representation(self, value):
        return value

    def to_internal_value(self, data):
        if isinstance(data, int):
            return data
        elif isinstance(data, str):
            data = data.upper()
            match = re.match(r'(\d+) *([MG])B*', data)
            if match:
                value = int(match.group(1))
                if match.group(2) == "M":
                    return value
                if match.group(2) == "G":
                    return value * 1024
                if match.group(2) == "K":
                    return value // 1024
            else:
                raise ValidationError("illegal size format")
        else:
            raise ValidationError("illegal type")


class IPListField(Field):
    def to_representation(self, value):
        return value

    def to_internal_value(self, data):
        try:
            [ip_address(ip) for ip in data.split(',')]
        except ValueError:
            raise ValidationError("not ipv4 address")
        return data



from typing import Dict, List
from operator import or_
from functools import reduce
from django.db.models import Q


def Q_any(key_list_maps: Dict[str, List]):
    return reduce(
        or_,
        (
            Q(**{key: value})
            for key in key_list_maps.keys()
            for value in key_list_maps[key]
        ),
        Q())



from werkzeug.routing import BaseConverter


class ListConverter(BaseConverter):
    def to_python(self, value):
        return value.split(',')

    def to_url(self, values):
        return ','.join(super().to_url(value)
                        for value in values)


class RegexConverter(BaseConverter):
    def __init__(self, map, *args):
        self.map = map
        self.regex = args[0]


class Converters:

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.url_map.converters['regex'] = RegexConverter
        app.url_map.converters['list'] = ListConverter



#!/bin/python3
from setuptools import setup
from setuptools import find_packages

NAME = "devtools"
PACKAGES = [NAME] + ["%s.%s" % (NAME, i) for i in find_packages(NAME)]
LONG_DESC = '''Some useful helper-funcs for devpers.
Sub-package is the extension of corresponding package with the same name.
'''


setup(
    name='AIJIdevtools',
    version='1.4.4',
    author='AIJI',
    author_email='thecrazyaiji@gmail.com',
    description='Some useful helper-funcs for devpers',
    long_description=LONG_DESC,
    packages=PACKAGES,
    install_requires=[
        'sh',
        'sqlparse',
        'termcolor',
        'requests',
    ],
    url='https://github.com/AIJIJI/devtools',
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 1 - Planning"
    ]
)



def partition(ls, size):
    """
    Returns a new list with elements
    of which is a list of certain size.

        >>> partition([1, 2, 3, 4], 3)
        [[1, 2, 3], [4]]
    """
    return [ls[i:i+size] for i in range(0, len(ls), size)]


def split(ls, sep=None):
    """
    Split a list by sep

        >>> split([1, None, None, 2, 3, None, 4])
        [[1], [2, 3], [4]]
    """
    res = []
    tmplist = []
    for elem in ls:
        if elem == sep:
            if tmplist:
                res.append(tmplist)
            tmplist = []
        else:
            tmplist.append(elem)
    else:
        res.append(tmplist)
    return res


def dict_ordered(ls, sep='\n', line_prefix=''):
    ls = sorted(ls)
    pre = ls[0][0]
    res = line_prefix + repr(ls[0]) + ','
    for s in ls[1:]:
        if s[0] == pre:
            res += ' ' + repr(s) + ','
        else:
            res += sep + line_prefix + repr(s) + ','
        pre = s[0]
    return res


if __name__ == '__main__':
    l = (
            'id', 'module_id', 'type', 'name', 'status_code',
            'status', 'flow_record', 'flow_nodes',
            'opinions', 'apply_type', 'manager', 'chinese_name',
            'app_level', 'desc', 'lang', 'gitlab', 'dependency_version',
            'basic_software', 'container', 'container_ver',
            'has_container', 'deploy_path', 'log_path',
            'script_path', 'heartbeat_path', 'running_params', 'port',
            'created_time', 'updated_time', 'app_id', 'names'
        )


    print('(\n'+dict_ordered(l, line_prefix=' '*12)+'\n'+' '*8+')')

#     l = '''

# '''
#     l = l.split('\n\n\n')
#     print('\n\n\n'.join(sorted(l)))